//
//
//  Real Danger
//
//  You controlling a little rabbit who is in real danger from a bunch of eagles
//  soaring above. Dodge them by moving left and right. As they get hungrier, 
//  the eagles also get faster. When you get caught, press A to try again.
//
//  If you have any questions regarding the code, contact me
//  via the method specified at  http://www.sebastianmihai.com
//  (operational in 2012), and I will try my best to help.
//
//                             Sebastian Mihai, 2012
//


//Include library, charmap, charset
#include <libgccvb.h>
#include "charset.h"
#include "bgmap.h"

//  Virtual Boy resolution is 384 x 224

// tiles
#define EAGLE_LEFT_1 3
#define EAGLE_LEFT_2 4
#define EAGLE_RIGHT_1 5
#define EAGLE_RIGHT_2 6
#define EAGLE_RISING_1 0
#define EAGLE_RISING_2 1
#define EAGLE_DIVING 2
#define RABBIT_LEFT_1 7
#define RABBIT_LEFT_2 8
#define RABBIT_RIGHT_1 9
#define RABBIT_RIGHT_2 10

#define MAX_EAGLES 8
#define RABBIT_SPEED 3

#define STATE_FLYING 0
#define STATE_DROPPING 1
#define STATE_RISING 2

u16 EAGLE_FLYING_SPEED;
u16 EAGLE_DROPPING_SPEED;
u16 EAGLE_RISING_SPEED;

u16 eagleX[ MAX_EAGLES ];
s16 eagleDX[ MAX_EAGLES ];
u16 eagleY[ MAX_EAGLES ];
u16 eagleState[ MAX_EAGLES ];
s16 rabbitCaptured;

// sprite (object) allocation
// 1023 - rabbit
// 1022 - first eagle
// 1021 - second eagle, and so on


// a solid pseudorandom function
WORD random( u16 frame ) {
	
	return ((frame * 14577) % 190) + 25; // homemade random number generator
}
 
void SetSpriteTile( u16 sprite, u16 tile ) {
	OAM[(sprite << 2) + 3] = (OAM[(sprite << 2) + 3] & 0xF000) | tile;
}

void SetSpritePosition( u16 sprite, u16 x, u16 y ) {
	OAM[ sprite << 2] = x;
	OAM[(sprite << 2) + 2] = y;
}

void animateRabbit( u16 x, u16 y, u16 frame, u16 rabbitDirection ) {
	if( frame % 10 < 5 ) {
		if( rabbitDirection == 0 )
			SetSpriteTile( 1023, RABBIT_LEFT_1 );
		else 
			SetSpriteTile( 1023, RABBIT_RIGHT_1 );
	} else {
		if( rabbitDirection == 0 )
			SetSpriteTile( 1023, RABBIT_LEFT_2 );
		else
			SetSpriteTile( 1023, RABBIT_RIGHT_2 );
	}
	
	SetSpritePosition( 1023, x, y );
}

void animateEagle( u16 which, u16 frame ) {
	if( eagleState[ which ] == STATE_FLYING ) {
		if( frame % 10 < 5 ) {
			if( eagleDX[ which ] < 0 )
				SetSpriteTile( 1022 - which, EAGLE_LEFT_1 );
			else
				SetSpriteTile( 1022 - which, EAGLE_RIGHT_1 );
		} else {
			if( eagleDX[ which ] < 0 )
				SetSpriteTile( 1022 - which, EAGLE_LEFT_2 );
			else
				SetSpriteTile( 1022 - which, EAGLE_RIGHT_2 );
		}
	}
	
	if( eagleState[ which ] == STATE_DROPPING ) {
		SetSpriteTile( 1022 - which, EAGLE_DIVING );
	}
	
	if( eagleState[ which ] == STATE_RISING ) {
		if( frame % 10 < 5 ) {
			SetSpriteTile( 1022 - which, EAGLE_RISING_1 );
		} else {
			SetSpriteTile( 1022 - which, EAGLE_RISING_2 );
		}
	}
	
	SetSpritePosition( 1022 - which, eagleX[ which ], eagleY[ which ] );
}

// attempt to move an eagle
void moveEagle( u16 which, u16 frame ) {
	// since we are moving all eagles based on the same frame, we need to add
	// something to frame to make the random number it generates a bit more random between eagles :)
	u16 r = random( frame + 37 * which );
	
	// if rabbit has been captured and this eagle does not have him, hide this eagle
	if( rabbitCaptured != -1 && rabbitCaptured != which ) {
		// offscreen
		eagleX[ which ] = 385;
		eagleY[ which ] = 225;
		return;
	}
	
	if( eagleState[ which ] == STATE_FLYING ) {
		// a small chance to keep dive, unless rabbit was captured
		if( r % 100 < 2 && rabbitCaptured == -1 ) {
			eagleState[ which ] = STATE_DROPPING;
		} else {
			r = random( frame + 253 * which );
			// keep flying, either left or right
			
			// small chance of turning around, unless rabbit was captured
			if( r % 100 < 1 && rabbitCaptured == -1 ) {
				eagleDX[ which ] = eagleDX[ which ] * (-1);
			}
			
			// do we need to turn around because we have reached the end of the screen?
			// right side?
			if( eagleX[ which ] + eagleDX[ which ] > 384 - 8 ) {
				eagleDX[ which ] = eagleDX[ which ] * (-1);
			}
			// left side?
			if( eagleX[ which ] + eagleDX[ which ] < 0 ) {
				eagleDX[ which ] = eagleDX[ which ] * (-1);
			}
			
			// fly!
			eagleX[ which ] = eagleX[ which ] + eagleDX[ which ];
		}
	} 
	
	if( eagleState[ which ] == STATE_DROPPING ) {
		// lower eagle until he reaches bottom of screen
		if( eagleY[ which ] < 224 - 8 - EAGLE_DROPPING_SPEED ) {
			eagleY[ which ] += EAGLE_DROPPING_SPEED;
		} else {
			eagleState[ which ] = STATE_RISING;
		}
	}
	
	if( eagleState[ which ] == STATE_RISING ) {
		// raise eagle until he reaches top of screen
		if( eagleY[ which ] > 0 + EAGLE_RISING_SPEED ) {
			eagleY[ which ] -= EAGLE_RISING_SPEED;
		} else {
			eagleState[ which ] = STATE_FLYING;
		}
	}
}

void initializeSprites() {
	u16 i;

	for (i=1023; i>=1022 - MAX_EAGLES; i--) {
		vbSetObject(i, OBJ_ON, (1023-i)*8, 0, 0, 0);
	}
	
	SetSpriteTile( 1023, RABBIT_LEFT_1 );
	
	for (i=0; i<MAX_EAGLES; i++) {
		
		SetSpriteTile( 1022 - i*4 - 0, EAGLE_LEFT_1 );
	}
}

void initializeEagles() {
	
	u16 i;
	
	EAGLE_FLYING_SPEED = 2;
	EAGLE_DROPPING_SPEED = 8;
	EAGLE_RISING_SPEED = 4;
	
	for (i=0; i<MAX_EAGLES; i++) {
		// spread out the eagles
		eagleX[ i ] = (random( i ) % 188) * 2;
		eagleY[ i ] = (i%2) * 8;
		eagleState[ i ] = STATE_FLYING;
		eagleDX[ i ] = EAGLE_FLYING_SPEED;
		if( i%2 == 0 ) 
			eagleDX[ i ] = eagleDX[ i ] * (-1);
	}
}

int game()
{
	u16 i;
	u16 x = 188; u16 y = 216;
	u16 joystick;
	u16 currentFrame = 0;
	u16 rabbitDirection = 0; // left

	rabbitCaptured = -1;
	
    //Initial setup
    vbDisplayOn();

    //Copy tileset and tilemap into memory
    copymem((void*)CharSeg0, (void*)CHADJUSTMENT, 256*16);
    copymem((void*)BGMap(0), (void*)BGADJUSTMENT, 450*16);

	VIP_REGS[SPT3]=1023;
    VIP_REGS[SPT2]=1015;

	initializeEagles();
	// associate tiles with sprites
    initializeSprites();

	vbSetWorld(31, WRLD_ON | WRLD_OBJ, 0, 0, 0, 0, 0, 0, 0, 0);
    vbSetWorld(30, WRLD_END, 0, 0, 0, 0, 0, 0, 0, 0);

    //Set brightness registers
    vbDisplayShow();
	
	while(1){
		currentFrame++;
		
		BGM_CSET(0, 15, 0, 0);
		BGM_PALSET(0,15,0,BGM_PAL3);
		
		joystick = vbReadPad();
		
		if( rabbitCaptured == -1 ) {
			if( joystick & K_LR ) {
				x+=RABBIT_SPEED;
				rabbitDirection = 1;
			} else if( joystick & K_LL ) {
				x-=RABBIT_SPEED;
				rabbitDirection = 0;
			}
		} else {
			// rabbit is being carried away... by eagle with index stored in rabbitCaptured
			x = eagleX[ rabbitCaptured ];
			y = eagleY[ rabbitCaptured ] + 6;
		}
		
		animateRabbit( x, y, currentFrame, rabbitDirection );
		
		for(i=0;i<MAX_EAGLES;i++) {
			moveEagle( i, currentFrame );
			animateEagle( i, currentFrame );
			
			if( rabbitCaptured == -1 ) {
				// has this eagle captured the rabbit?
				if( (  ( eagleX[ i ] > x && eagleX[ i ] - x < 8 ) ||
					 ( eagleX[ i ] < x && x - eagleX[ i ] < 8 ) ) 
				&&
					( y - eagleY[ i ] < 12 )
				) {
					rabbitCaptured = i;
				}
			}
		}
		
		// constantly speed up eagles, unless rabbit was caught
		if( currentFrame % 200 == 0 && rabbitCaptured == -1 ) {
			EAGLE_FLYING_SPEED++;
			EAGLE_DROPPING_SPEED++;
			EAGLE_RISING_SPEED++;
			
			for(i=0;i<MAX_EAGLES;i++) {
				if( eagleDX[ i ] < 0 )
					eagleDX[ i ] = EAGLE_FLYING_SPEED * (-1);
				else 
					eagleDX[ i ] = EAGLE_FLYING_SPEED;
			}
		}
		
		// wait for a press and release of the A button if the rabbit has been caught
		if( rabbitCaptured != -1 ) {
			if( joystick & K_A ) {
				while( vbReadPad() & K_A ) {
				}
				return 0;
			}
		}
		
		vbWaitFrame(1);
		
	}

    return 0;
}

int main() {
	while(1) {
		game();
	}
}